"""Network - Security Groups

This module contains actions for viewing & managing Azure Network security groups.
"""

from typing import List, Optional, Union

from ppa import ui

from .exceptions import InvalidRule
from ._network_api import client, process_security_group, process_security_groups
from .._common import input_table as common_input_table
from .records import SecurityGroup, SecurityGroupRule
from ..records import AzureClient
from azure.mgmt.network.models import SecurityRule, SecurityRuleAccess, SecurityRuleDirection, SecurityRuleProtocol


def get_all(azure_client: AzureClient, subscription_id: str, resource_group_name: Optional[str] = None) -> List[SecurityGroup]:
    """Get all security groups.

    !!! question "Minimum Plugin Version: 1.1.0"

    ??? info "Input"

        - __azure_client:__ an [AzureClient](../inputs_outputs/#azureclient) dictionary

        - __subscription_id__: the subscription ID

        - __resource_group_name:__ the resource group name (Optional)

    ??? success "Output"
        A list of [SecurityGroup](../inputs_outputs/#securitygroup) dictionaries.

    ??? example
        ```YAML linenums="1"
        - azure.network.security_groups.get_all:
          load:
            azure_client: azure_secrets
            subscription_id: subscription_id
          save: all_security_groups
        ```
    """
    return process_security_groups(
        client(client=azure_client, subscription_id=subscription_id).network_security_groups.list(resource_group_name)
        if resource_group_name is not None
        else client(client=azure_client, subscription_id=subscription_id).network_security_groups.list_all()
    )


def get_by_name(azure_client: AzureClient, subscription_id: str, name: str, resource_group_name: str) -> SecurityGroup:
    """Get a security group with a specific name.

    !!! question "Minimum Plugin Version: 1.1.0"

    ??? info "Input"

        - __azure_client:__ an [AzureClient](../inputs_outputs/#azureclient) dictionary

        - __subscription_id__: the subscription ID

        - __name:__ the security group name

        - __resource_group_name:__ the resource group name

    ??? success "Output"
        A [SecurityGroup](../inputs_outputs/#securitygroup) dictionary.

    ??? example
        ```YAML linenums="1"
        - azure.network.security_groups.get_by_name:
            name: ppa-dev
            resource_group_name: ppa
          load:
            azure_client: azure_secrets
            subscription_id: subscription_id
          save: selected_group
        ```
    """
    return process_security_group(
        client(client=azure_client, subscription_id=subscription_id).network_security_groups.get(resource_group_name, name)
    )


GROUP_TABLE_HEADER = ["Name", "ID", "Location", "Tags", "Number of Security Rules"]

GROUP_TABLE_FIELDS = [
    lambda group: group.get("name"),
    lambda group: group.get("id"),
    lambda group: group.get("location"),
    lambda group: group.get("tags") or "",
    lambda group: len(group.get("security_rules")),
]


def input_table(
    text: str, groups: List[SecurityGroup], minimum: Optional[int] = None, maximum: Optional[int] = None
) -> List[SecurityGroup]:
    """Display a list of instances in a table, & allow the task operator to make a selection.

    The table will have the following columns:

    - Name
    - ID
    - Location
    - Tags
    - Number of Security Rules

    !!! question "Minimum Plugin Version: 1.1.0"

    ??? info "Input"

        - __text:__ the title of the table

        - __groups:__ a list of [SecurityGroup](../inputs_outputs/#securitygroup) dictionaries to display in the table

        - __minimum:__ The minimum number of acceptable selections (Optional)

        - __maximum:__ The maximum number of acceptable selections (Optional)

    ??? success "Output"

        A list of [SecurityGroup](../inputs_outputs/#securitygroup) dictionaries.

    ??? tip

        - If __neither__ a `minimum` _or_ `maximum` is provided, the task operator will be able to submit 0 selections.

        - If `minimum` _or_ `maximum` are provided, the operation will repeat until a valid number of selections is made.

    ??? example

        Finding groups with [get_all](#get_all), saving them as `all_groups`, & waiting for a single selection:

        ```YAML linenums="1" hl_lines="5 12 13"
        - azure.network.security_groups.get_all:
          load:
            subscription_id: subscription_id
            azure_client: azure_secrets
          save: all_groups

        - azure.network.security_groups.input_table:
            text: Choose a security group
            minimum: 1
            maximum: 1
          load:
            groups: all_groups
          save: selected_groups
        ```
    """

    return common_input_table(
        text=text, data=groups, minimum=minimum, maximum=maximum, header=GROUP_TABLE_HEADER, fields=GROUP_TABLE_FIELDS
    )


def output_table(text: str, groups: Union[SecurityGroup, List[SecurityGroup]]) -> None:
    """Display a list of security groups in a table.

    The table will have the following columns:

    - Name
    - ID
    - Location
    - Tags
    - Number of Security Rules

    !!! question "Minimum Plugin Version: 1.1.0"

    ??? info "Input"

        - __text:__ the title of the table

        - __groups:__ a single or list of [SecurityGroup](../inputs_outputs/#securitygroup) dictionaries

    ??? success "Output"

        Nothing is outputted by this action.

    ??? example

        Finding all groups in a resource group with [get_all](#get_all), saving them as `all_groups`, & displaying them:

        ```YAML linenums="1" hl_lines="5 10"
        - azure.network.security_groups.get_all:
            resource_group_name: ppa
          load:
            subscription_id: subscription_id
            azure_client: azure_secrets
          save: all_groups

        - azure.network.security_groups.output_table:
            text: All Groups
          load:
            groups: all_groups
        ```
    """
    if not isinstance(groups, list):
        groups = [groups]

    ui.output_table(text, header=GROUP_TABLE_HEADER, data=[[get(group) for get in GROUP_TABLE_FIELDS] for group in groups])


RULES_TABLE_HEADER = [
    "Name",
    "Description",
    "Protocol",
    "Direction",
    "Access",
    "Priority",
    "Source Port Range",
    "Source Address Prefix",
    "Source Address Prefixes",
    "Destination Port Range",
    "Destination Address Prefix",
    "Destination Address Prefixes",
]

RULES_TABLE_FIELDS = [
    lambda rule_set: rule_set["name"] or "",
    lambda rule_set: rule_set["description"] or "",
    lambda rule_set: rule_set["protocol"],
    lambda rule_set: rule_set["direction"],
    lambda rule_set: rule_set["access"],
    lambda rule_set: rule_set["priority"],
    lambda rule_set: rule_set["source_port_range"],
    lambda rule_set: rule_set["source_address_prefix"] or "",
    lambda rule_set: rule_set["source_address_prefixes"] or "",
    lambda rule_set: rule_set["destination_port_range"],
    lambda rule_set: rule_set["destination_address_prefix"] or "",
    lambda rule_set: rule_set["destination_address_prefixes"] or "",
]


def input_security_rules_table(
    text: str, group: SecurityGroup, minimum: Optional[int] = None, maximum: Optional[int] = None
) -> List[SecurityGroupRule]:
    """Display a list of security rules in a table, & allow the task operator to make a selection.

    The table will have the following columns:

    - Name
    - Description
    - Protocol
    - Direction
    - Access
    - Priority
    - Source Port Range
    - Source Address Prefix
    - Source Address Prefixes
    - Destination Port Range
    - Destination Address Prefix
    - Destination Address Prefixes

    !!! question "Minimum Plugin Version: 1.1.0"

    ??? info "Input"

        - __text:__ the title of the table

        - __group:__ a [SecurityGroup](../inputs_outputs/#securitygroup) dictionary

        - __minimum:__ The minimum number of acceptable selections (Optional)

        - __maximum:__ The maximum number of acceptable selections (Optional)

    ??? success "Output"

        A list of [SecurityGroupRule](../inputs_outputs/#securitygrouprule) dictionaries.

    ??? tip

        - If __neither__ a `minimum` _or_ `maximum` is provided, the task operator will be able to submit 0 selections.

        - If `minimum` _or_ `maximum` are provided, the operation will repeat until a valid number of selections is made.

    ??? example

        Finding groups with [get_all](#get_all), saving them as `all_groups`, & waiting for a single selection:

        ```YAML linenums="1" hl_lines="5 12 13"
        - azure.network.security_groups.get_by_name:
            name: ppa-dev
            resource_group_name: ppa
          load:
            azure_client: azure_secrets
            subscription_id: subscription_id
          save: group

        - azure.network.security_groups.input_security_rules_table:
            text: Choose a security rule
            minimum: 1
            maximum: 1
          load:
            group: group
          save: selected_rules
        ```
    """

    return common_input_table(
        text=text,
        data=group["security_rules"] or [],
        minimum=minimum,
        maximum=maximum,
        header=RULES_TABLE_HEADER,
        fields=RULES_TABLE_FIELDS,
    )


def output_security_rules_table(text: str, group: SecurityGroup) -> None:
    """Display the security rules for a security group in a table.

    The table will have the following columns:

    - Name
    - Description
    - Protocol
    - Direction
    - Access
    - Priority
    - Source Port Range
    - Source Address Prefix
    - Source Address Prefixes
    - Destination Port Range
    - Destination Address Prefix
    - Destination Address Prefixes

    !!! question "Minimum Plugin Version: 1.1.0"

    ??? info "Input"

        - __text:__ the title of the table

        - __group:__ a single [SecurityGroup](../inputs_outputs/#securitygroup) dictionary

    ??? success "Output"

        Nothing is outputted by this action.

    ??? example

        Finding a security group with [get_by_name](#get_by_name), saving it as `group`, & displaying the inbound rules:

        ```YAML linenums="1" hl_lines="6 11"
        - azure.network.security_groups.get_by_name:
            name: ppa-dev
            resource_group_name: ppa
          load:
            azure_client: azure_secrets
            subscription_id: subscription_id
          save: group

        - azure.network.security_groups.output_security_rules_table:
            text: Security rules for ppa-dev
          load:
            group: group
        ```
    """

    ui.output_table(
        text,
        header=RULES_TABLE_HEADER,
        data=[[get(rule_set) for get in RULES_TABLE_FIELDS] for rule_set in group["security_rules"]],
    )


def create_rule(
    azure_client: AzureClient,
    subscription_id: str,
    resource_group_name: str,
    name: str,
    access: str,
    rule_name: str,
    protocol: str,
    direction: str,
    source_port_range: str,
    destination_port_range: str,
    source_address_prefix: str,
    destination_address_prefix: str,
    priority: int,
    rule_description: Optional[str] = "",
) -> None:
    """Add a security rule to a security group.

    !!! question "Minimum Plugin Version: 1.1.0"

    !!! tip "Idempotent Action"

        This action will do nothing if a matching rule already exists in the security group.

    ??? info "Input"

        - __azure_client:__ an [AzureClient](../inputs_outputs/#azureclient) dictionary

        - __subscription_id__: the subscription ID

        - __name:__ the security group name

        - __resource_group_name:__ the resource group name

        - __rule_name__: the rule name

        - __access:__ `allow` or `deny`

        - __protocol:__ `tcp`, `udp`, or `*` for any

        - __direction:__ `inbound` or `outbound`

        - __source_port_range:__ provide a single port or port range

        - __destination_port_range:__ provide a single port or port range

        - __source_address_prefix:__ an IP address range

        - __destination_address_prefix:__ an IP address range

        - __priority:__ rules are processed in priority order; the lower the number, the higher the priority

        - __rule_description:__ description of the rule (optional)


    ??? success "Output"

        Nothing is outputted by this action.

    ??? example

        ```YAML linenums="1"
        - azure.network.security_groups.create_rule:
            protocol: tcp
            direction: inbound
            source_port_range: "*"
            destination_address_prefix: "*"
            name: ppa-dev
            resource_group_name: ppa
            rule_name: inbound rule 80
            rule_description: this is the description
            access: allow
            priority: 200
            destination_port_range: 80
          load:
            source_address_prefix: source_address
            azure_client: azure_secrets
            subscription_id: subscription_id
        ```
    """

    if access not in ["allow", "deny"]:
        raise InvalidRule("access must be 'deny' or 'allow'. Provided value: {}".format(access))

    if direction not in ["inbound", "outbound"]:
        raise InvalidRule("direction must be 'inbound' or 'outbound'. Provided value: {}".format(direction))

    if protocol not in ["tcp", "udp", "*"]:
        raise InvalidRule("protocol must be 'tcp', 'udp', or '*'. Provided value: {}".format(protocol))

    client(client=azure_client, subscription_id=subscription_id).security_rules.create_or_update(
        resource_group_name,
        name,
        rule_name,
        SecurityRule(
            name=rule_name,
            access=SecurityRuleAccess[access].value,
            description=rule_description,
            destination_address_prefix=destination_address_prefix,
            destination_port_range=destination_port_range,
            direction=SecurityRuleDirection[direction].value,
            priority=priority,
            protocol=SecurityRuleProtocol[protocol].value,
            source_address_prefix=source_address_prefix,
            source_port_range=source_port_range,
        ),
    ).result()


def delete_rule(azure_client: AzureClient, subscription_id: str, resource_group_name: str, name: str, rule_name: str,) -> None:
    """Delete a security rule from a security group.

    !!! question "Minimum Plugin Version: 1.1.0"

    ??? info "Input"

        - __azure_client:__ an [AzureClient](../inputs_outputs/#azureclient) dictionary

        - __subscription_id__: the subscription ID

        - __name:__ the security group name

        - __resource_group_name:__ the resource group name

        - __rule_name__: the rule name

    ??? success "Output"

        Nothing is outputted by this action.

    ??? example

        ```YAML linenums="1"

        - azure.network.security_groups.delete_rule:
            name: ppa-dev
            resource_group_name: ppa
            rule_name: inbound 80
          load:
            azure_client: azure_secrets
            subscription_id: subscription_id
        ```
    """
    client(client=azure_client, subscription_id=subscription_id).security_rules.delete(
        resource_group_name, name, rule_name,
    )
