"""Storage - Files

This module contains actions for uploading/downloading files to/from an Azure Storage File Share.
"""
import logging
import pathlib
import contextlib

from ppa import ui

from azure.core.exceptions import ClientAuthenticationError, ResourceNotFoundError
from azure.storage.fileshare import ShareFileClient


SHARED_WORKSPACE = pathlib.Path("/workspace/share/")

logger = logging.getLogger(__name__)


class AuthenticationFailed(Exception):
    pass


@contextlib.contextmanager
def _files_api(connection_string, file_share_name, target):
    try:
        yield ShareFileClient.from_connection_string(conn_str=connection_string, share_name=file_share_name, file_path=target)
    except ClientAuthenticationError as e:
        logger.error(str(e))
        message = f"Authentication to file share {file_share_name} failed, please check the connection string."
        ui.output_error(message)
        raise AuthenticationFailed(message)
    except (ResourceNotFoundError, FileNotFoundError) as e:
        # Always raise FileNotFound so the calling function only has to catch one type of exception.
        logger.error(str(e))
        raise FileNotFoundError()
    except ValueError as e:
        if "either blank or malformed" in str(e):
            message = (
                f"Authentication to file share {file_share_name} failed as the connection string is not in the correct format."
            )
            ui.output_error(message)
            raise AuthenticationFailed(message)
        raise e


def download(connection_string: str, file_share_name: str, source: str, destination: str) -> None:
    """Download a file from an Azure Storage File Share to the task workspace.

    !!! question "Minimum Plugin Version: 3.1.0"

    ??? info "Input"

        - __connection_string:__ an Azure Storage Account connection string (obtained from the Azure Portal)

        - __file_share_name__: the name of the File Share in the Azure Storage Account

        - __source:__ the _full path_ to the file in the File Share (including extension)

        - __destination:__ the _name_ of the file to create in the task workspace (including extension)

    ??? success "Output"

        Nothing is outputted by this action.

    ??? example
        ```YAML linenums="1"
        - azure.storage.files.download:
            file_share_name: ppa-file-share
            source: example-file.txt
            destination: example-file.txt
          load:
            connection_string: connection_string
        ```
    """
    logger.info(f"downloading file from {source} to destination {destination}...")
    try:
        with _files_api(connection_string, file_share_name, source) as client:
            with (SHARED_WORKSPACE / destination).open("wb+") as open_file:
                client.download_file().readinto(open_file)
    except FileNotFoundError:
        message = f"The download action failed as file '{source}' does not exist in file share {file_share_name}."
        ui.output_error(message)
        raise FileNotFoundError(message)


def upload(connection_string: str, file_share_name: str, source: str, destination: str) -> None:
    """Upload a file from the task workspace to an Azure Storage File Share.

    !!! warning "If the target file already exists it will be overwritten."

    !!! question "Minimum Plugin Version: 3.1.0"

    ??? info "Input"

        - __connection_string:__ an Azure Storage Account connection string (obtained from the Azure Portal)

        - __file_share_name__: the name of the File Share in the Azure Storage Account

        - __source:__ the _name_ of the file to upload from the task workspace (including extension)

        - __destination:__ the _full path_ to the file's destination in the File Share (including extension)

    ??? success "Output"

        Nothing is outputted by this action.

    ??? example
        ```YAML linenums="1"
        - azure.storage.files.upload:
            file_share_name: ppa-file-share
            source: example-file.txt
            destination: example-file.txt
          load:
            connection_string: connection_string
        ```
    """
    logger.info(f"uploading file from {source} to destination {destination}...")
    try:
        with _files_api(connection_string, file_share_name, destination) as client:
            with (SHARED_WORKSPACE / source).open("rb") as open_file:
                client.upload_file(open_file)
    except FileNotFoundError:
        message = f"The upload action failed as file '{source}' does not exist in the task workspace."
        ui.output_error(message)
        raise FileNotFoundError(message)


def delete(connection_string: str, file_share_name: str, file_path: str) -> None:
    """Delete a file from an Azure Storage File Share.

The action will succeed if the file does not exist.

    !!! question "Minimum Plugin Version: 3.1.0"

    ??? info "Input"

        - __connection_string:__ an Azure Storage Account connection string (obtained from the Azure Portal)

        - __file_share_name__: the name of the File Share in the Azure Storage Account

        - __file_path:__ the _full path_ to the target file in the File Share (including extension)

    ??? success "Output"

        Nothing is outputted by this action.

    ??? example
        ```YAML linenums="1"
        - azure.storage.files.delete:
            file_share_name: ppa-file-share
            file_path: example-file.txt
          load:
            connection_string: connection_string
        ```
    """
    logger.info(f"deleting file {file_path} from {file_share_name}...")
    try:
        with _files_api(connection_string, file_share_name, file_path) as client:
            client.delete_file()
    except FileNotFoundError:
        logger.info(f"cannot delete file {file_path} on file share {file_share_name} as it does not exist")
